/*
 * Written by Sinan & Sami.  Copyright 1993, TUBITAK Software
 * Research & Development Center, Ankara, TURKIYE
 * This program is freely distributable without licensing fees and
 * is provided without guarantee or warrantee expressed or implied.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "type.h"

/*
 ******  global variables
*/

vertex *TOKEN_TABLE[MAX_NO_OF_TOKENS];
unsigned NO_OF_TOKENS=0,MAX_LEVEL=0;
char fName[30];

struct queue{
    unsigned item;
    struct queue *child;
};
struct queue *frontPointer,*rearPointer;

/*
 **** function error()
*/

void error()
{
    printf("insufficient memory\n");
    exit(1);
}

/*
 *** function getTokens()
 *** given a string it returns the token names in an array called tokens
 *** which can contain up to 50 tokens
 *** tokens are separated by commas.
*/

void getTokens(char *buf,char tokens[50][NAMELENGTH],unsigned *noOfTokens)
{
 char *s;
 (*noOfTokens)=0;
 if (!(s=strtok(buf,":"))) return;
 strcpy(tokens[0],s);

 while ((s=strtok(NULL,","))!=NULL)
 {
   (*noOfTokens)++;
   strcpy(tokens[(*noOfTokens)],s);
 }

  if ((s=strtok(buf,"\0"))!=NULL){
      (*noOfTokens)++;
      strcpy(tokens[(*noOfTokens)],s);
  }
}


/*
 **** function foundInTable()
 **** if the token is in the TOKEN_TABLE, returns its id
 **** otherwise returns 0.
*/

int foundInTable(char *token)
{ unsigned i;
  for (i=1;i<=NO_OF_TOKENS;i++)
    if (strcmp(TOKEN_TABLE[i]->name,token)==0)
       return(i);                               /* if found return id */
  return(0);                                    /* if not found  return 0*/
}




/*
 **** function insertIntoTable()
 **** inserts a new token to the TOKEN_TABLE
*/

void insertIntoTable(char *token)
{
  NO_OF_TOKENS++;
  if ((TOKEN_TABLE[NO_OF_TOKENS]=(vertex*) malloc(sizeof(vertex)))==NULL)
     error();

  strcpy(TOKEN_TABLE[NO_OF_TOKENS]->name,token);
  TOKEN_TABLE[NO_OF_TOKENS]->id=NO_OF_TOKENS;
  TOKEN_TABLE[NO_OF_TOKENS]->level=0;
  TOKEN_TABLE[NO_OF_TOKENS]->no_of_child=0;
  TOKEN_TABLE[NO_OF_TOKENS]->no_of_parent=0;
}





/*
 **** function parsing()
 **** opens the input stream, reads the statements separated by semicolons
 **** and by calling getTokens(), foundInTable(), and insertIntoTable()
 **** functions forms the TOKEN_TABLE
*/

void parsing()
{
   FILE *stream;
   char buf[255],ch,tokens[50][NAMELENGTH];
   unsigned noOfTokens,i;
   unsigned parentNo;

   if ((stream = fopen(fName, "rt"))
      == NULL)
   {
     printf("Cannot open input file.\n");
     exit(1);
   }

   while (!feof(stream))
   {
     i=0;

     do{
       fread(&ch,1,1,stream);
       if (ch!='\n') buf[i++]=ch;
     } while ((ch!=';') && (!feof(stream)));

     buf[i-1]=0;

     getTokens(buf,tokens,&noOfTokens);

     if (noOfTokens>0)
     {
	for (i=0;i<noOfTokens;i++)
	  if (!foundInTable(tokens[i]))
	      insertIntoTable(tokens[i]);

	parentNo=foundInTable(tokens[0]);
	if (noOfTokens>1)
	{
	   TOKEN_TABLE[parentNo]->child =
	       (unsigned *)  malloc((noOfTokens-1)*sizeof(unsigned));

	   if (TOKEN_TABLE[parentNo]->child == NULL)
	       error();

	   for (i=1;i<noOfTokens;i++)
	   {
	      TOKEN_TABLE[parentNo]->child[i-1]=foundInTable(tokens[i]);
	      TOKEN_TABLE[parentNo]->no_of_child++;
	   }
	 } /* if    */
     }     /* if    */
   }       /* while */

   fclose(stream);
}


/*
  **** function reArrange()
  **** rearranges the TOKEN_TABLE so that the first entry is a dummy token
  **** whose children is the level 1 tokens.
*/

void reArrange()
{
  unsigned i,j,k,found,counter=0;
  unsigned tokens[50];

  if ((TOKEN_TABLE[0]=(vertex*) malloc(sizeof(vertex)))==NULL)
      error();

  for (i=1;i<=NO_OF_TOKENS;i++){
    found=0;
    for (j=1;(j<=NO_OF_TOKENS) && !found;j++)
      for (k=0;k<TOKEN_TABLE[j]->no_of_child;k++)
	if (TOKEN_TABLE[j]->child[k]==i) found=1;
    if (!found)
       tokens[counter++]=i;
  }

  TOKEN_TABLE[0]->child=(unsigned *) malloc(counter*sizeof(unsigned));
  if (TOKEN_TABLE[0]->child == NULL)
     error();

  for (i=0;i<counter;i++)
     TOKEN_TABLE[0]->child[i]=tokens[i];

  TOKEN_TABLE[0]->no_of_child=counter;
  TOKEN_TABLE[0]->no_of_parent=0;
  TOKEN_TABLE[0]->level=0;
  TOKEN_TABLE[0]->id=0;
  strcpy(TOKEN_TABLE[0]->name,"DUMMY NAME");
}



/*
 **** function insertQ()
 **** inserts a new item to the queue called from function BFS()
*/

void insertQ(unsigned v)
{
   struct queue *p=(struct queue *) malloc(sizeof(struct queue));
   p->item=v;
   p->child=NULL;
   rearPointer->child=p;
   rearPointer=rearPointer->child;
}



/*
 **** function removeQ()
 **** removes an item from the queue called from function BFS()
*/

unsigned removeQ()
{
  struct queue *p=frontPointer;
  frontPointer=p->child;
  free(p);
  return(frontPointer->item);
}



/*
 **** function EmptyQ()
 **** tests whether the queue is empty, called from function BFS()
*/

int emptyQ()
{
  return(frontPointer==rearPointer ? 1:0);
}



/*
 **** function foundInQ()
 **** returns 1 if the item is already in queue otherwise returns 0
*/

int foundInQ(unsigned no)
{
  struct queue *p=frontPointer;
  while (p!=rearPointer)
  {
    if (p->item==no)
	return 1;           /* found */
    p=p->child;
  }
  return 0;   /* not found */
}




/*
 **** function BFS()
 **** finds the levels of each entry in the TOKEN_TABLE
*/

void BFS()
{ unsigned FLAG=50000;
  unsigned counter=0,i;
  unsigned v=0;
  unsigned po=TOKEN_TABLE[v]->no_of_child;
  unsigned o=0,t=0;

 frontPointer=rearPointer=(struct queue *) malloc(sizeof(struct queue));


  insertQ(FLAG);
  insertQ(v);
  insertQ(FLAG);

  while (!emptyQ())
  {
    v=removeQ();
      if (v==FLAG) {counter++;continue;}
    for (i=0;i<TOKEN_TABLE[v]->no_of_child;i++)
    {
      o++;
      t+=TOKEN_TABLE[TOKEN_TABLE[v]->child[i]]->no_of_child;
      if (TOKEN_TABLE[TOKEN_TABLE[v]->child[i]]->level<counter)
          TOKEN_TABLE[TOKEN_TABLE[v]->child[i]]->level=counter;

          insertQ(TOKEN_TABLE[v]->child[i]);
        if (o==po)
        {
            insertQ(FLAG);
            o=0;
            po=t;
            t=0;
        }

    }  /* for   */
  }    /* while */

  MAX_LEVEL=counter-1;
  free(frontPointer);
  free(rearPointer);
}

/*
 **** function dummynodes()
 ****
*/

void dummynodes()
{
  unsigned i,j,no,no_of_child;
  for (i=1;i<=NO_OF_TOKENS;i++)
  {
    no_of_child=TOKEN_TABLE[i]->no_of_child;
    for (j=0;j<no_of_child;j++)
    {
      no = TOKEN_TABLE[i]->child[j];
      if ((TOKEN_TABLE[no]->level - TOKEN_TABLE[i]->level) > 1)
      {
	insertIntoTable("du");
	TOKEN_TABLE[NO_OF_TOKENS]->level = TOKEN_TABLE[i]->level+1;
	TOKEN_TABLE[NO_OF_TOKENS]->child=(unsigned *) malloc(sizeof(unsigned));
	TOKEN_TABLE[NO_OF_TOKENS]->child[0]=no;
	TOKEN_TABLE[NO_OF_TOKENS]->no_of_child = 1;
	TOKEN_TABLE[NO_OF_TOKENS]->parent=(unsigned *) malloc(sizeof(unsigned));
	TOKEN_TABLE[NO_OF_TOKENS]->parent[0]=TOKEN_TABLE[i]->child[j];
	TOKEN_TABLE[NO_OF_TOKENS]->no_of_parent = 1;

	TOKEN_TABLE[i]->child[j] = NO_OF_TOKENS;
      }
    }
  }
}



/*
 **** function parents()
 ****
*/

void parents()
{
  typedef struct {
     unsigned parents[50];
     unsigned noOfParents;
   } fat;

  unsigned i,j,no;

  fat *fathers=(fat *) malloc(NO_OF_TOKENS*sizeof(fat));
  if (fathers==NULL)
     error();

  /** initialize the array **/

  for (i=0;i<NO_OF_TOKENS;i++)
     fathers[i].noOfParents=0;

  /** determine the parents of the vertexes **/

  for (i=1;i<=NO_OF_TOKENS;i++)
    for (j=0;j<TOKEN_TABLE[i]->no_of_child;j++)
    {
       no=TOKEN_TABLE[i]->child[j]-1;
       fathers[no].parents[fathers[no].noOfParents++]=i;
    }

  for (i=0;i<NO_OF_TOKENS;i++)
    if (fathers[i].noOfParents>0)
    {
       TOKEN_TABLE[i+1]->parent =
	     (unsigned *) malloc(fathers[i].noOfParents*sizeof(unsigned));
       if (TOKEN_TABLE[i+1]->parent == NULL)
	  error();
       TOKEN_TABLE[i+1]->no_of_parent = fathers[i].noOfParents;
       for (j=0;j<fathers[i].noOfParents;j++)
	 TOKEN_TABLE[i+1]->parent[j] = fathers[i].parents[j];
    }
}





/*
 **** function main()
 ****
*/

int levels(int argc, char *argv[])
{
 extern vertex **tree;
 unsigned i,j,counter;

 if (argc>1)
    strcpy(fName,argv[1]);
 else
    strcpy(fName,"names.dat");

 parsing();
 reArrange();
 BFS();
 dummynodes();
 parents();

 tree=(vertex **) malloc((MAX_LEVEL+2)*sizeof(vertex *));

 for (i=1;i<=MAX_LEVEL;i++)
 {
   counter=0;
   for (j=1;j<=NO_OF_TOKENS;j++)
     if (TOKEN_TABLE[j]->level==i)
       counter++;

   tree[i-1] = (vertex *) malloc((counter+1)*sizeof(vertex));

   if (tree[i-1]==NULL)
      error();

   counter=0;
   for (j=1;j<=NO_OF_TOKENS;j++)
      if (TOKEN_TABLE[j]->level==i)
      {
	strcpy(tree[i-1][counter].name,TOKEN_TABLE[j]->name);
	tree[i-1][counter].id=TOKEN_TABLE[j]->id;
	tree[i-1][counter].level=TOKEN_TABLE[j]->level-1;
	tree[i-1][counter].no_of_child=TOKEN_TABLE[j]->no_of_child;
	tree[i-1][counter].no_of_parent=TOKEN_TABLE[j]->no_of_parent;
	tree[i-1][counter].child=TOKEN_TABLE[j]->child;
	tree[i-1][counter].parent=TOKEN_TABLE[j]->parent;
	counter++;
      }

   tree[i-1][counter].id=0;
 }
  tree[MAX_LEVEL]=(vertex *)malloc(2*sizeof(vertex));
  tree[MAX_LEVEL][0].id=0;

  i=0;
  while (tree[i][0].id!=0)
  {
    j=0;
    while (tree[i][j].id!=0)
    {
      tree[i][j].baryUp=0;
      tree[i][j].baryDown=0;
      j++;
    }
    i++;
  }


 /*** remove the token table from memory
					*/
 for (i=0;i<=NO_OF_TOKENS;i++)
    free(TOKEN_TABLE[i]);

 return(0);

}


