/*
 * Written by Sinan & Sami.  Copyright 1993, TUBITAK Software
 * Research & Development Center, Ankara, TURKIYE
 * This program is freely distributable without licensing fees and
 * is provided without guarantee or warrantee expressed or implied.
 *
 */

#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "type.h"

int down;
int doubleCheck;
int noOfLevels=4;
int levelCross(vertex * a);
vertex **tree;
#define BARY(x)     (down ? x->baryDown : x->baryUp)
#define UN_BARY(x)  (down ? x->baryUp   : x->baryDown)


int levelLength(vertex *a)
{
    int i;
    for(i=0;a[i].id!=0;i++);
    return(i);
}

int medianComp(const void *a, const void *b)
{
   unsigned *x=(unsigned *)a;
   unsigned *y=(unsigned *)b;

   if(*x == *y)
     return(0);

   if(*x > *y)
     return(1);
   return(-1);
}

unsigned * intchr(unsigned *a,unsigned b,unsigned c)
{
  int i=0;
    while(i<c)
     {
       if(b==a[i])
	 return(a+i);
	 i++;
     }
  return(NULL);
}

int compareValue(const void *a, const void *b)
{
   vertex *x=(vertex *)a;
   vertex *y=(vertex *)b;
   if(BARY(x) == BARY(y)){
     if(UN_BARY(x) > UN_BARY(y))
	 return(1);
     if(UN_BARY(x) < UN_BARY(y))
	 return(-1);
     return 0;
   }
   if(BARY(x) > BARY(y))
     return(1);
   return(-1);
}

void  medianValue(vertex *a)
{
    int i,j;
    unsigned left,right;
    float m;
    unsigned orders[80];
    vertex *adjacent;
    adjacent=tree[down ? (a->level)+1 : (a->level)-1];

    #define RELATIVE(x) (down ? x->child : x->parent)
    #define LENGTH(x)   (down ? x->no_of_child : x->no_of_parent)

    for(j=0,i=1; i<=levelLength(adjacent) ; i++)
       if(intchr(RELATIVE(a),adjacent[i-1].id,LENGTH(a)))
       {
	   orders[j]=i;
	   j++;
       }
    qsort(orders,j,sizeof(unsigned),medianComp);


    if (j==0)
    {
       if(down) a->baryDown = -1.0; else a->baryUp = -1.0;
       return;
    }
    if(j%2)
    {
       if(down)a->baryDown=orders[(j-1)/2]; else a->baryUp = orders[(j-1)/2];
       return;
    }
    if(j==2)
    {
       if(down)
           a->baryDown = (orders[0]+orders[1])/2.0;
       else
           a->baryUp = (orders[0]+orders[1])/2.0;
       return;
    }
    left =orders[j/2-1] -orders[0];
    right=orders[j-1]   -orders[j/2];

    if( down )
        a->baryDown=1.0*(orders[j/2-1]*right + 1.0*orders[j/2]*left)/(left+right);
    else
        a->baryUp=1.0*(orders[j/2-1]*right + 1.0*orders[j/2]*left)/(left+right);

}
int equals(vertex * a, int t)
{
  int res=0,i;
  vertex temp;
  for( i=0;i<t-1;i++)

    if(BARY((a+i)) == BARY((a+i+1)))
    {
       memcpy(&temp,a+i  ,sizeof(vertex));
       memcpy(a+i  ,a+i+1,sizeof(vertex));
       memcpy(a+i+1,&temp,sizeof(vertex));
       res=1;
    }
return(res);
}
void medianSort(vertex* a,vertex *b)
{
    int t,i,p;
    vertex* dummy,*temp;
    t=levelLength(a);
    dummy=(vertex*)malloc((t+1)*sizeof(vertex));
    down = a[0].level > b[0].level ? FALSE : TRUE;

    for(i=0;i<t;i++)
	 medianValue(a+i);
    memcpy(dummy,a,(t+1)*sizeof(vertex));

    qsort(a,t,sizeof(vertex),compareValue);

    if(levelCross(a)>levelCross(dummy))
    {
      temp=a;
      tree[a[0].level]=dummy;
      a=dummy;
      dummy=temp;
    }

    for( p=0;p<5;p++)
    {
	memcpy(dummy,a,(t+1)*sizeof(vertex));

	if(!equals(a,t)) break;

	if(levelCross(a)>=levelCross(dummy))
	{
	  temp=a;
	  tree[a[0].level]=dummy;
	  a=dummy;
	  dummy=temp;
	}
    }

    free(dummy);
}

int twoVertCross(vertex *a,vertex *b,unsigned *sorted,unsigned k)
{
   unsigned i,j,sum=0;
   unsigned *c,*d;
   for(i=0;i<a->no_of_child;i++)
      for(j=0;j<b->no_of_child;j++)
	 {
	    c=intchr(sorted,a->child[i],k);
	    d=intchr(sorted,b->child[j],k);
	    if((c-sorted)>(d-sorted))
	      sum++;
	 }
  return sum;
}

int levelCross(vertex * a)
{
   unsigned i,j,t,k,sum=0;
   vertex *b;
   unsigned *sorted;

   if(a[0].level && (!down || doubleCheck))
   {
      b=tree[a[0].level-1];
      k=levelLength(a);
      sorted=(unsigned *)malloc((k+2)*sizeof(unsigned));
      for(i=0;i<k;i++)
	 sorted[i]=a[i].id;
      t=levelLength(b);
      for(i=0;i<t-1;i++)
	 for(j=i+1;j<t;j++)
	   sum+=twoVertCross(b+i,b+j,sorted,k);
      free(sorted);
   }

   if((a[0].level < noOfLevels-1) && (down || doubleCheck))
   {
      b=tree[a[0].level+1];
      k=levelLength(b);
      sorted=(unsigned *)malloc((k+2)*sizeof(unsigned));
      for(i=0;i<k;i++)
	 sorted[i]=b[i].id;
      t=levelLength(a);
      for(i=0;i<t-1;i++)
	 for(j=i+1;j<t;j++)
	   sum+=twoVertCross(a+i,a+j,sorted,k);
      free(sorted);
   }

   return(sum);
}


int median(int argc, char *argv[])
{
    int i,j,k;
    levels(argc,argv);
    i=0;
    while(tree[i][0].id != 0)
       i++;

    for(k=0;k<15;k++)
    {
       doubleCheck=(k%4)*((k+1)%4);
       for(j=1;j<i;j++)
       {
	  medianSort(tree[j],tree[j-1]);
       }
       for(j=i-1;j>0;j--)
       {
	  medianSort(tree[j-1],tree[j]);
       }

       for(j=1;j<i;j++)
       {
	  medianSort(tree[j-1],tree[j]);
       }
       for(j=i-1;j>0;j--)
       {
	  medianSort(tree[j],tree[j-1]);
       }
    }

    for(i=0;tree[i][0].id;i++)
    {
      j=0;
      while(tree[i][j].id)
      {
	 tree[i][j].y=i*150;
	 tree[i][j].x=(j+1)*1000/(levelLength(tree[i]))+50;
         j++;
      }
    }
}

